@@ -1,18 +1,7 @@
 //
-// Copyright 2014-15 Ettus Research LLC
+// Copyright 2014-17 Ettus Research, A National Instruments Company
 //
-// This program is free software: you can redistribute it and/or modify
-// it under the terms of the GNU General Public License as published by
-// the Free Software Foundation, either version 3 of the License, or
-// (at your option) any later version.
-//
-// This program is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-// GNU General Public License for more details.
-//
-// You should have received a copy of the GNU General Public License
-// along with this program.  If not, see <http://www.gnu.org/licenses/>.
+// SPDX-License-Identifier: GPL-3.0-or-later
 //
 
 /***********************************************************************
@@ -28,6 +17,9 @@
 #include <uhd/utils/assert_has.hpp>
 #include <uhd/utils/log.hpp>
 #include <uhd/utils/static.hpp>
+#include <uhd/utils/safe_call.hpp>
+#include <uhdlib/usrp/common/max287x.hpp>
+
 #include <boost/assign/list_of.hpp>
 #include <boost/shared_ptr.hpp>
 #include <boost/math/special_functions/round.hpp>
@@ -35,7 +27,6 @@
 #include <boost/algorithm/string.hpp>
 #include <boost/thread/mutex.hpp>
 #include <map>
-#include "max287x.hpp"
 
 using namespace uhd;
 using namespace uhd::usrp;
@@ -86,7 +77,8 @@ enum ubx_cpld_field_id_t
     RXDRV_FORCEON = 21,
     RXAMP_FORCEON = 22,
     RXLNA1_FORCEON = 23,
-    RXLNA2_FORCEON = 24
+    RXLNA2_FORCEON = 24,
+    CAL_ENABLE = 25
 };
 
 struct ubx_gpio_field_info_t
@@ -158,6 +150,8 @@ static const dboard_id_t UBX_V2_160MHZ_TX_ID(0x7D);
 static const dboard_id_t UBX_V2_160MHZ_RX_ID(0x7E);
 static const dboard_id_t UBX_LP_160MHZ_TX_ID(0x0200);
 static const dboard_id_t UBX_LP_160MHZ_RX_ID(0x0201);
+static const dboard_id_t UBX_TDD_160MHZ_TX_ID(0x0202);
+static const dboard_id_t UBX_TDD_160MHZ_RX_ID(0x0203);
 static const freq_range_t ubx_freq_range(10e6, 6.0e9);
 static const gain_range_t ubx_tx_gain_range(0, 31.5, double(0.5));
 static const gain_range_t ubx_rx_gain_range(0, 31.5, double(0.5));
@@ -225,6 +219,14 @@ public:
         _iface = get_iface();
         dboard_id_t rx_id = get_rx_id();
         dboard_id_t tx_id = get_tx_id();
+        size_t revision = 1;    // default to rev A
+        // Get revision if programmed
+        const std::string revision_str = get_rx_eeprom().revision;
+        if (not revision_str.empty())
+        {
+            revision = boost::lexical_cast<size_t>(revision_str);
+        }
+        _high_isolation = false;
         if (rx_id == UBX_PROTO_V3_RX_ID and tx_id == UBX_PROTO_V3_TX_ID) {
             _rev = 0;
         }
@@ -236,6 +238,10 @@ public:
         }
         else if (rx_id == UBX_V2_40MHZ_RX_ID and tx_id == UBX_V2_40MHZ_TX_ID) {
             _rev = 2;
+            if (revision >= 4)
+            {
+                _high_isolation = true;
+            }
         }
         else if (rx_id == UBX_V1_160MHZ_RX_ID and tx_id == UBX_V1_160MHZ_TX_ID) {
             bw = 160e6;
@@ -244,12 +250,21 @@ public:
         else if (rx_id == UBX_V2_160MHZ_RX_ID and tx_id == UBX_V2_160MHZ_TX_ID) {
             bw = 160e6;
             _rev = 2;
+            if (revision >= 4)
+            {
+                _high_isolation = true;
+            }
         }
         else if (rx_id == UBX_LP_160MHZ_RX_ID and tx_id == UBX_LP_160MHZ_TX_ID) {
             // The LP version behaves and looks like a regular UBX-160 v2
             bw = 160e6;
             _rev = 2;
         }
+        else if (rx_id == UBX_TDD_160MHZ_RX_ID and tx_id == UBX_TDD_160MHZ_TX_ID) {
+            bw = 160e6;
+            _rev = 2;
+            _high_isolation = true;
+        }
         else {
             UHD_THROW_INVALID_CODE_PATH();
         }
@@ -299,7 +314,7 @@ public:
             {
                 std::vector<double> rates = _iface->get_clock_rates(dboard_iface::UNIT_RX);
                 double highest_rate = 0.0;
-                BOOST_FOREACH(double rate, rates)
+                for(double rate:  rates)
                 {
                     if (rate <= pfd_freq_max and rate > highest_rate)
                         highest_rate = rate;
@@ -307,7 +322,9 @@ public:
                 try {
                     _iface->set_clock_rate(dboard_iface::UNIT_RX, highest_rate);
                 } catch (const uhd::not_implemented_error &) {
-                    UHD_MSG(warning) << "Unable to set dboard clock rate - phase will vary" << std::endl;
+                    UHD_LOG_WARNING("UBX",
+                        "Unable to set dboard clock rate - phase will vary"
+                    );
                     can_set_clock_rate = false;
                 }
                 _rx_target_pfd_freq = highest_rate;
@@ -316,7 +333,7 @@ public:
             {
                 std::vector<double> rates = _iface->get_clock_rates(dboard_iface::UNIT_TX);
                 double highest_rate = 0.0;
-                BOOST_FOREACH(double rate, rates)
+                for(double rate:  rates)
                 {
                     if (rate <= pfd_freq_max and rate > highest_rate)
                         highest_rate = rate;
@@ -324,7 +341,9 @@ public:
                 try {
                     _iface->set_clock_rate(dboard_iface::UNIT_TX, highest_rate);
                 } catch (const uhd::not_implemented_error &) {
-                    UHD_MSG(warning) << "Unable to set dboard clock rate - phase will vary" << std::endl;
+                    UHD_LOG_WARNING("UBX",
+                        "Unable to set dboard clock rate - phase will vary"
+                    );
                 }
                 _tx_target_pfd_freq = highest_rate;
             }
@@ -351,14 +370,14 @@ public:
         write_gpio();
 
         // Configure ATR
-        _iface->set_atr_reg(dboard_iface::UNIT_TX, dboard_iface::ATR_REG_IDLE, _tx_gpio_reg.atr_idle);
-        _iface->set_atr_reg(dboard_iface::UNIT_TX, dboard_iface::ATR_REG_TX_ONLY, _tx_gpio_reg.atr_tx);
-        _iface->set_atr_reg(dboard_iface::UNIT_TX, dboard_iface::ATR_REG_RX_ONLY, _tx_gpio_reg.atr_rx);
-        _iface->set_atr_reg(dboard_iface::UNIT_TX, dboard_iface::ATR_REG_FULL_DUPLEX, _tx_gpio_reg.atr_full_duplex);
-        _iface->set_atr_reg(dboard_iface::UNIT_RX, dboard_iface::ATR_REG_IDLE, _rx_gpio_reg.atr_idle);
-        _iface->set_atr_reg(dboard_iface::UNIT_RX, dboard_iface::ATR_REG_TX_ONLY, _rx_gpio_reg.atr_tx);
-        _iface->set_atr_reg(dboard_iface::UNIT_RX, dboard_iface::ATR_REG_RX_ONLY, _rx_gpio_reg.atr_rx);
-        _iface->set_atr_reg(dboard_iface::UNIT_RX, dboard_iface::ATR_REG_FULL_DUPLEX, _rx_gpio_reg.atr_full_duplex);
+        _iface->set_atr_reg(dboard_iface::UNIT_TX, gpio_atr::ATR_REG_IDLE, _tx_gpio_reg.atr_idle);
+        _iface->set_atr_reg(dboard_iface::UNIT_TX, gpio_atr::ATR_REG_TX_ONLY, _tx_gpio_reg.atr_tx);
+        _iface->set_atr_reg(dboard_iface::UNIT_TX, gpio_atr::ATR_REG_RX_ONLY, _tx_gpio_reg.atr_rx);
+        _iface->set_atr_reg(dboard_iface::UNIT_TX, gpio_atr::ATR_REG_FULL_DUPLEX, _tx_gpio_reg.atr_full_duplex);
+        _iface->set_atr_reg(dboard_iface::UNIT_RX, gpio_atr::ATR_REG_IDLE, _rx_gpio_reg.atr_idle);
+        _iface->set_atr_reg(dboard_iface::UNIT_RX, gpio_atr::ATR_REG_TX_ONLY, _rx_gpio_reg.atr_tx);
+        _iface->set_atr_reg(dboard_iface::UNIT_RX, gpio_atr::ATR_REG_RX_ONLY, _rx_gpio_reg.atr_rx);
+        _iface->set_atr_reg(dboard_iface::UNIT_RX, gpio_atr::ATR_REG_FULL_DUPLEX, _rx_gpio_reg.atr_full_duplex);
 
         // Engage ATR control (1 is ATR control, 0 is manual control)
         _iface->set_pin_ctrl(dboard_iface::UNIT_TX, _tx_gpio_reg.atr_mask);
@@ -378,7 +397,7 @@ public:
             _rxlo1 = max287x_iface::make<max2870>(boost::bind(&ubx_xcvr::write_spi_regs, this, RXLO1, _1));
             _rxlo2 = max287x_iface::make<max2870>(boost::bind(&ubx_xcvr::write_spi_regs, this, RXLO2, _1));
             std::vector<max287x_iface::sptr> los = boost::assign::list_of(_txlo1)(_txlo2)(_rxlo1)(_rxlo2);
-            BOOST_FOREACH(max287x_iface::sptr lo, los)
+            for(max287x_iface::sptr lo:  los)
             {
                 lo->set_auto_retune(false);
                 lo->set_muxout_mode(max287x_iface::MUXOUT_DLD);
@@ -392,7 +411,7 @@ public:
             _rxlo1 = max287x_iface::make<max2871>(boost::bind(&ubx_xcvr::write_spi_regs, this, RXLO1, _1));
             _rxlo2 = max287x_iface::make<max2871>(boost::bind(&ubx_xcvr::write_spi_regs, this, RXLO2, _1));
             std::vector<max287x_iface::sptr> los = boost::assign::list_of(_txlo1)(_txlo2)(_rxlo1)(_rxlo2);
-            BOOST_FOREACH(max287x_iface::sptr lo, los)
+            for(max287x_iface::sptr lo:  los)
             {
                 lo->set_auto_retune(false);
                 //lo->set_cycle_slip_mode(true);  // tried it - caused longer lock times
@@ -417,23 +436,23 @@ public:
         get_rx_subtree()->create<std::vector<std::string> >("power_mode/options")
             .set(ubx_power_modes);
         get_rx_subtree()->create<std::string>("power_mode/value")
-            .subscribe(boost::bind(&ubx_xcvr::set_power_mode, this, _1))
+            .add_coerced_subscriber(boost::bind(&ubx_xcvr::set_power_mode, this, _1))
             .set("performance");
         get_rx_subtree()->create<std::vector<std::string> >("xcvr_mode/options")
             .set(ubx_xcvr_modes);
         get_rx_subtree()->create<std::string>("xcvr_mode/value")
-            .subscribe(boost::bind(&ubx_xcvr::set_xcvr_mode, this, _1))
+            .add_coerced_subscriber(boost::bind(&ubx_xcvr::set_xcvr_mode, this, _1))
             .set("FDX");
         get_tx_subtree()->create<std::vector<std::string> >("power_mode/options")
             .set(ubx_power_modes);
         get_tx_subtree()->create<std::string>("power_mode/value")
-            .subscribe(boost::bind(&uhd::property<std::string>::set, &get_rx_subtree()->access<std::string>("power_mode/value"), _1))
-            .publish(boost::bind(&uhd::property<std::string>::get, &get_rx_subtree()->access<std::string>("power_mode/value")));
+            .add_coerced_subscriber(boost::bind(&uhd::property<std::string>::set, &get_rx_subtree()->access<std::string>("power_mode/value"), _1))
+            .set_publisher(boost::bind(&uhd::property<std::string>::get, &get_rx_subtree()->access<std::string>("power_mode/value")));
         get_tx_subtree()->create<std::vector<std::string> >("xcvr_mode/options")
             .set(ubx_xcvr_modes);
         get_tx_subtree()->create<std::string>("xcvr_mode/value")
-            .subscribe(boost::bind(&uhd::property<std::string>::set, &get_rx_subtree()->access<std::string>("xcvr_mode/value"), _1))
-            .publish(boost::bind(&uhd::property<std::string>::get, &get_rx_subtree()->access<std::string>("xcvr_mode/value")));
+            .add_coerced_subscriber(boost::bind(&uhd::property<std::string>::set, &get_rx_subtree()->access<std::string>("xcvr_mode/value"), _1))
+            .set_publisher(boost::bind(&uhd::property<std::string>::get, &get_rx_subtree()->access<std::string>("xcvr_mode/value")));
 
         ////////////////////////////////////////////////////////////////////
         // Register TX properties
@@ -442,20 +461,20 @@ public:
         get_tx_subtree()->create<device_addr_t>("tune_args")
             .set(device_addr_t());
         get_tx_subtree()->create<sensor_value_t>("sensors/lo_locked")
-            .publish(boost::bind(&ubx_xcvr::get_locked, this, "TXLO"));
+            .set_publisher(boost::bind(&ubx_xcvr::get_locked, this, "TXLO"));
         get_tx_subtree()->create<double>("gains/PGA0/value")
-            .coerce(boost::bind(&ubx_xcvr::set_tx_gain, this, _1)).set(0);
+            .set_coercer(boost::bind(&ubx_xcvr::set_tx_gain, this, _1)).set(0);
         get_tx_subtree()->create<meta_range_t>("gains/PGA0/range")
             .set(ubx_tx_gain_range);
         get_tx_subtree()->create<double>("freq/value")
-            .coerce(boost::bind(&ubx_xcvr::set_tx_freq, this, _1))
+            .set_coercer(boost::bind(&ubx_xcvr::set_tx_freq, this, _1))
             .set(ubx_freq_range.start());
         get_tx_subtree()->create<meta_range_t>("freq/range")
             .set(ubx_freq_range);
         get_tx_subtree()->create<std::vector<std::string> >("antenna/options")
             .set(ubx_tx_antennas);
         get_tx_subtree()->create<std::string>("antenna/value")
-            .subscribe(boost::bind(&ubx_xcvr::set_tx_ant, this, _1))
+            .add_coerced_subscriber(boost::bind(&ubx_xcvr::set_tx_ant, this, _1))
             .set(ubx_tx_antennas.at(0));
         get_tx_subtree()->create<std::string>("connection")
             .set("QI");
@@ -468,7 +487,7 @@ public:
         get_tx_subtree()->create<meta_range_t>("bandwidth/range")
             .set(freq_range_t(bw, bw));
         get_tx_subtree()->create<int64_t>("sync_delay")
-            .subscribe(boost::bind(&ubx_xcvr::set_sync_delay, this, true, _1))
+            .add_coerced_subscriber(boost::bind(&ubx_xcvr::set_sync_delay, this, true, _1))
             .set(-8);
 
         ////////////////////////////////////////////////////////////////////
@@ -478,21 +497,21 @@ public:
         get_rx_subtree()->create<device_addr_t>("tune_args")
             .set(device_addr_t());
         get_rx_subtree()->create<sensor_value_t>("sensors/lo_locked")
-            .publish(boost::bind(&ubx_xcvr::get_locked, this, "RXLO"));
+            .set_publisher(boost::bind(&ubx_xcvr::get_locked, this, "RXLO"));
         get_rx_subtree()->create<double>("gains/PGA0/value")
-            .coerce(boost::bind(&ubx_xcvr::set_rx_gain, this, _1))
+            .set_coercer(boost::bind(&ubx_xcvr::set_rx_gain, this, _1))
             .set(0);
         get_rx_subtree()->create<meta_range_t>("gains/PGA0/range")
             .set(ubx_rx_gain_range);
         get_rx_subtree()->create<double>("freq/value")
-            .coerce(boost::bind(&ubx_xcvr::set_rx_freq, this, _1))
+            .set_coercer(boost::bind(&ubx_xcvr::set_rx_freq, this, _1))
             .set(ubx_freq_range.start());
         get_rx_subtree()->create<meta_range_t>("freq/range")
             .set(ubx_freq_range);
         get_rx_subtree()->create<std::vector<std::string> >("antenna/options")
             .set(ubx_rx_antennas);
         get_rx_subtree()->create<std::string>("antenna/value")
-            .subscribe(boost::bind(&ubx_xcvr::set_rx_ant, this, _1)).set("RX2");
+            .add_coerced_subscriber(boost::bind(&ubx_xcvr::set_rx_ant, this, _1)).set("RX2");
         get_rx_subtree()->create<std::string>("connection")
             .set("IQ");
         get_rx_subtree()->create<bool>("enabled")
@@ -504,39 +523,43 @@ public:
         get_rx_subtree()->create<meta_range_t>("bandwidth/range")
             .set(freq_range_t(bw, bw));
         get_rx_subtree()->create<int64_t>("sync_delay")
-            .subscribe(boost::bind(&ubx_xcvr::set_sync_delay, this, false, _1))
+            .add_coerced_subscriber(boost::bind(&ubx_xcvr::set_sync_delay, this, false, _1))
             .set(-8);
     }
 
-    ~ubx_xcvr(void)
+    virtual ~ubx_xcvr(void)
     {
-        // Shutdown synthesizers
-        _txlo1->shutdown();
-        _txlo2->shutdown();
-        _rxlo1->shutdown();
-        _rxlo2->shutdown();
+        UHD_SAFE_CALL
+        (
+            // Shutdown synthesizers
+            _txlo1->shutdown();
+            _txlo2->shutdown();
+            _rxlo1->shutdown();
+            _rxlo2->shutdown();
 
-        // Reset CPLD values
-        _cpld_reg.value = 0;
-        write_cpld_reg();
+            // Reset CPLD values
+            _cpld_reg.value = 0;
+            write_cpld_reg();
 
-        // Reset GPIO values
-        set_gpio_field(TX_GAIN, 0);
-        set_gpio_field(CPLD_RST_N, 0);
-        set_gpio_field(RX_ANT, 1);
-        set_gpio_field(TX_EN_N, 1);
-        set_gpio_field(RX_EN_N, 1);
-        set_gpio_field(SPI_ADDR, 0x7);
-        set_gpio_field(RX_GAIN, 0);
-        set_gpio_field(TXLO1_SYNC, 0);
-        set_gpio_field(TXLO2_SYNC, 0);
-        set_gpio_field(RXLO1_SYNC, 0);
-        set_gpio_field(RXLO1_SYNC, 0);
-        write_gpio();
+            // Reset GPIO values
+            set_gpio_field(TX_GAIN, 0);
+            set_gpio_field(CPLD_RST_N, 0);
+            set_gpio_field(RX_ANT, 1);
+            set_gpio_field(TX_EN_N, 1);
+            set_gpio_field(RX_EN_N, 1);
+            set_gpio_field(SPI_ADDR, 0x7);
+            set_gpio_field(RX_GAIN, 0);
+            set_gpio_field(TXLO1_SYNC, 0);
+            set_gpio_field(TXLO2_SYNC, 0);
+            set_gpio_field(RXLO1_SYNC, 0);
+            set_gpio_field(RXLO1_SYNC, 0);
+            write_gpio();
+        )
     }
 
 private:
     enum power_mode_t {PERFORMANCE,POWERSAVE};
+    enum xcvr_mode_t {FDX, TDD, TX, RX, FAST_TDD};
 
     /***********************************************************************
     * Helper Functions
@@ -552,7 +575,7 @@ private:
     {
         boost::mutex::scoped_lock lock(_spi_mutex);
         ROUTE_SPI(_iface, dest);
-        BOOST_FOREACH(uint32_t value, values)
+        for(uint32_t value:  values)
             WRITE_SPI(_iface, value);
     }
 
@@ -670,23 +693,23 @@ private:
             uint16_t mask = lo1_field_info.mask | lo2_field_info.mask;
             dboard_iface::unit_t unit = lo1_field_info.unit;
             UHD_ASSERT_THROW(lo1_field_info.unit == lo2_field_info.unit);
-            _iface->set_atr_reg(unit, dboard_iface::ATR_REG_IDLE, value, mask);
+            _iface->set_atr_reg(unit, gpio_atr::ATR_REG_IDLE, value, mask);
             cmd_time += uhd::time_spec_t(1/pfd_freq);
             _iface->set_command_time(cmd_time);
-            _iface->set_atr_reg(unit, dboard_iface::ATR_REG_TX_ONLY, value, mask);
+            _iface->set_atr_reg(unit, gpio_atr::ATR_REG_TX_ONLY, value, mask);
             cmd_time += uhd::time_spec_t(1/pfd_freq);
             _iface->set_command_time(cmd_time);
-            _iface->set_atr_reg(unit, dboard_iface::ATR_REG_RX_ONLY, value, mask);
+            _iface->set_atr_reg(unit, gpio_atr::ATR_REG_RX_ONLY, value, mask);
             cmd_time += uhd::time_spec_t(1/pfd_freq);
             _iface->set_command_time(cmd_time);
-            _iface->set_atr_reg(unit, dboard_iface::ATR_REG_FULL_DUPLEX, value, mask);
+            _iface->set_atr_reg(unit, gpio_atr::ATR_REG_FULL_DUPLEX, value, mask);
 
             // De-assert SYNC
             // Head of line blocking means the command time does not need to be set.
-            _iface->set_atr_reg(unit, dboard_iface::ATR_REG_IDLE, 0, mask);
-            _iface->set_atr_reg(unit, dboard_iface::ATR_REG_TX_ONLY, 0, mask);
-            _iface->set_atr_reg(unit, dboard_iface::ATR_REG_RX_ONLY, 0, mask);
-            _iface->set_atr_reg(unit, dboard_iface::ATR_REG_FULL_DUPLEX, 0, mask);
+            _iface->set_atr_reg(unit, gpio_atr::ATR_REG_IDLE, 0, mask);
+            _iface->set_atr_reg(unit, gpio_atr::ATR_REG_TX_ONLY, 0, mask);
+            _iface->set_atr_reg(unit, gpio_atr::ATR_REG_RX_ONLY, 0, mask);
+            _iface->set_atr_reg(unit, gpio_atr::ATR_REG_FULL_DUPLEX, 0, mask);
         }
     }
 
@@ -716,6 +739,8 @@ private:
     {
         //validate input
         assert_has(ubx_tx_antennas, ant, "ubx tx antenna name");
+        set_cpld_field(CAL_ENABLE, (ant == "CAL"));
+        write_cpld_reg();
     }
 
     // Set RX antennas
@@ -725,19 +750,20 @@ private:
         //validate input
         assert_has(ubx_rx_antennas, ant, "ubx rx antenna name");
 
-        // Due to an issue with TX path into to the RF switch (U32), there
-        // is a long transient at the beginning of transmission when the RX
-        // antenna is set to RX2.  Forcing on the TX PA removes the transient,
-        // so it is forced on only when the RX2 antenna is selected.  It is
-        // cleared when the TX/RX antenna is selected to avoid a higher noise
-        // floor on RX.
+        // There can be long transients on TX, so force on the TX PA
+        // except when in powersave mode (to save power) or on early
+        // boards that had lower TX-RX isolation when the RX antenna
+        // is set to TX/RX (to prevent higher noise floor on RX).
+        // Setting the xcvr_mode to TDD will force on the PA when
+        // not in powersave mode regardless of the board revision.
         if (ant == "TX/RX")
         {
             set_gpio_field(RX_ANT, 0);
-            set_cpld_field(TXDRV_FORCEON, 0);   // Turn off PA in TDD mode
+            // Force on TX PA for boards with high isolation or if the user sets the TDD mode
+            set_cpld_field(TXDRV_FORCEON, (_power_mode == POWERSAVE ? 0 : _high_isolation or _xcvr_mode == TDD ? 1 : 0));
         } else {
             set_gpio_field(RX_ANT, 1);
-            set_cpld_field(TXDRV_FORCEON, 1);   // Keep PA on
+            set_cpld_field(TXDRV_FORCEON, (_power_mode == POWERSAVE ? 0 : 1));   // Keep PA on
         }
         write_gpio();
         write_cpld_reg();
@@ -754,7 +780,7 @@ private:
         _ubx_tx_atten_val = ((attn_code & 0x3F) << 10);
         set_gpio_field(TX_GAIN, attn_code);
         write_gpio();
-        UHD_LOGV(rarely) << boost::format("UBX TX Gain: %f dB, Code: %d, IO Bits 0x%04x") % gain % attn_code % _ubx_tx_atten_val << std::endl;
+        UHD_LOGGER_TRACE("UBX") << boost::format("UBX TX Gain: %f dB, Code: %d, IO Bits 0x%04x") % gain % attn_code % _ubx_tx_atten_val ;
         _tx_gain = gain;
         return gain;
     }
@@ -767,7 +793,7 @@ private:
         _ubx_rx_atten_val = ((attn_code & 0x3F) << 10);
         set_gpio_field(RX_GAIN, attn_code);
         write_gpio();
-        UHD_LOGV(rarely) << boost::format("UBX RX Gain: %f dB, Code: %d, IO Bits 0x%04x") % gain % attn_code % _ubx_rx_atten_val << std::endl;
+        UHD_LOGGER_TRACE("UBX") << boost::format("UBX RX Gain: %f dB, Code: %d, IO Bits 0x%04x") % gain % attn_code % _ubx_rx_atten_val ;
         _rx_gain = gain;
         return gain;
     }
@@ -791,7 +817,21 @@ private:
         property_tree::sptr subtree = this->get_tx_subtree();
         device_addr_t tune_args = subtree->access<device_addr_t>("tune_args").get();
         is_int_n = boost::iequals(tune_args.get("mode_n",""), "integer");
-        UHD_LOGV(rarely) << boost::format("UBX TX: the requested frequency is %f MHz") % (freq/1e6) << std::endl;
+        UHD_LOGGER_TRACE("UBX") << boost::format("UBX TX: the requested frequency is %f MHz") % (freq/1e6) ;
+        double target_pfd_freq = _tx_target_pfd_freq;
+        if (is_int_n and tune_args.has_key("int_n_step"))
+        {
+            target_pfd_freq = tune_args.cast<double>("int_n_step", _tx_target_pfd_freq);
+            if (target_pfd_freq > _tx_target_pfd_freq)
+            {
+                UHD_LOGGER_WARNING("UBX")
+                    << boost::format("Requested int_n_step of %f MHz too large, clipping to %f MHz")
+                    % (target_pfd_freq/1e6)
+                    % (_tx_target_pfd_freq/1e6)
+                    ;
+                target_pfd_freq = _tx_target_pfd_freq;
+            }
+        }
 
         // Clip the frequency to the valid range
         freq = ubx_freq_range.clip(freq);
@@ -828,10 +868,10 @@ private:
             set_cpld_field(TXLB_SEL, 1);
             set_cpld_field(TXHB_SEL, 0);
             // Set LO1 to IF of 2100 MHz (offset from RX IF to reduce leakage)
-            freq_lo1 = _txlo1->set_frequency(2100*fMHz, ref_freq, _tx_target_pfd_freq, is_int_n);
+            freq_lo1 = _txlo1->set_frequency(2100*fMHz, ref_freq, target_pfd_freq, is_int_n);
             _txlo1->set_output_power(max287x_iface::OUTPUT_POWER_5DBM);
             // Set LO2 to IF minus desired frequency
-            freq_lo2 = _txlo2->set_frequency(freq_lo1 - freq, ref_freq, _tx_target_pfd_freq, is_int_n);
+            freq_lo2 = _txlo2->set_frequency(freq_lo1 - freq, ref_freq, target_pfd_freq, is_int_n);
             _txlo2->set_output_power(max287x_iface::OUTPUT_POWER_2DBM);
         }
         else if ((freq >= (500*fMHz)) && (freq <= (800*fMHz)))
@@ -841,7 +881,7 @@ private:
             set_cpld_field(TXLO1_FSEL1, 1);
             set_cpld_field(TXLB_SEL, 0);
             set_cpld_field(TXHB_SEL, 1);
-            freq_lo1 = _txlo1->set_frequency(freq, ref_freq, _tx_target_pfd_freq, is_int_n);
+            freq_lo1 = _txlo1->set_frequency(freq, ref_freq, target_pfd_freq, is_int_n);
             _txlo1->set_output_power(max287x_iface::OUTPUT_POWER_2DBM);
         }
         else if ((freq > (800*fMHz)) && (freq <= (1000*fMHz)))
@@ -851,7 +891,7 @@ private:
             set_cpld_field(TXLO1_FSEL1, 1);
             set_cpld_field(TXLB_SEL, 0);
             set_cpld_field(TXHB_SEL, 1);
-            freq_lo1 = _txlo1->set_frequency(freq, ref_freq, _tx_target_pfd_freq, is_int_n);
+            freq_lo1 = _txlo1->set_frequency(freq, ref_freq, target_pfd_freq, is_int_n);
             _txlo1->set_output_power(max287x_iface::OUTPUT_POWER_5DBM);
         }
         else if ((freq > (1000*fMHz)) && (freq <= (2200*fMHz)))
@@ -861,7 +901,7 @@ private:
             set_cpld_field(TXLO1_FSEL1, 0);
             set_cpld_field(TXLB_SEL, 0);
             set_cpld_field(TXHB_SEL, 1);
-            freq_lo1 = _txlo1->set_frequency(freq, ref_freq, _tx_target_pfd_freq, is_int_n);
+            freq_lo1 = _txlo1->set_frequency(freq, ref_freq, target_pfd_freq, is_int_n);
             _txlo1->set_output_power(max287x_iface::OUTPUT_POWER_2DBM);
         }
         else if ((freq > (2200*fMHz)) && (freq <= (2500*fMHz)))
@@ -871,7 +911,7 @@ private:
             set_cpld_field(TXLO1_FSEL1, 0);
             set_cpld_field(TXLB_SEL, 0);
             set_cpld_field(TXHB_SEL, 1);
-            freq_lo1 = _txlo1->set_frequency(freq, ref_freq, _tx_target_pfd_freq, is_int_n);
+            freq_lo1 = _txlo1->set_frequency(freq, ref_freq, target_pfd_freq, is_int_n);
             _txlo1->set_output_power(max287x_iface::OUTPUT_POWER_2DBM);
         }
         else if ((freq > (2500*fMHz)) && (freq <= (6000*fMHz)))
@@ -881,7 +921,7 @@ private:
             set_cpld_field(TXLO1_FSEL1, 0);
             set_cpld_field(TXLB_SEL, 0);
             set_cpld_field(TXHB_SEL, 1);
-            freq_lo1 = _txlo1->set_frequency(freq, ref_freq, _tx_target_pfd_freq, is_int_n);
+            freq_lo1 = _txlo1->set_frequency(freq, ref_freq, target_pfd_freq, is_int_n);
             _txlo1->set_output_power(max287x_iface::OUTPUT_POWER_5DBM);
         }
 
@@ -916,7 +956,7 @@ private:
         _txlo1_freq = freq_lo1;
         _txlo2_freq = freq_lo2;
 
-        UHD_LOGV(rarely) << boost::format("UBX TX: the actual frequency is %f MHz") % (_tx_freq/1e6) << std::endl;
+        UHD_LOGGER_TRACE("UBX") << boost::format("UBX TX: the actual frequency is %f MHz") % (_tx_freq/1e6) ;
 
         return _tx_freq;
     }
@@ -929,11 +969,25 @@ private:
         double ref_freq = _iface->get_clock_rate(dboard_iface::UNIT_RX);
         bool is_int_n = false;
 
-        UHD_LOGV(rarely) << boost::format("UBX RX: the requested frequency is %f MHz") % (freq/1e6) << std::endl;
+        UHD_LOGGER_TRACE("UBX") << boost::format("UBX RX: the requested frequency is %f MHz") % (freq/1e6) ;
 
         property_tree::sptr subtree = this->get_rx_subtree();
         device_addr_t tune_args = subtree->access<device_addr_t>("tune_args").get();
         is_int_n = boost::iequals(tune_args.get("mode_n",""), "integer");
+        double target_pfd_freq = _rx_target_pfd_freq;
+        if (is_int_n and tune_args.has_key("int_n_step"))
+        {
+            target_pfd_freq = tune_args.cast<double>("int_n_step", _rx_target_pfd_freq);
+            if (target_pfd_freq > _rx_target_pfd_freq)
+            {
+                UHD_LOGGER_WARNING("UBX")
+                    << boost::format("Requested int_n_step of %f Mhz too large, clipping to %f MHz")
+                    % (target_pfd_freq/1e6)
+                    % (_rx_target_pfd_freq/1e6)
+                    ;
+                target_pfd_freq = _rx_target_pfd_freq;
+            }
+        }
 
         // Clip the frequency to the valid range
         freq = ubx_freq_range.clip(freq);
@@ -972,10 +1026,10 @@ private:
             set_cpld_field(RXLB_SEL, 1);
             set_cpld_field(RXHB_SEL, 0);
             // Set LO1 to IF of 2380 MHz (2440 MHz filter center minus 60 MHz offset to minimize LO leakage)
-            freq_lo1 = _rxlo1->set_frequency(2380*fMHz, ref_freq, _rx_target_pfd_freq, is_int_n);
+            freq_lo1 = _rxlo1->set_frequency(2380*fMHz, ref_freq, target_pfd_freq, is_int_n);
             _rxlo1->set_output_power(max287x_iface::OUTPUT_POWER_5DBM);
             // Set LO2 to IF minus desired frequency
-            freq_lo2 = _rxlo2->set_frequency(freq_lo1 - freq, ref_freq, _rx_target_pfd_freq, is_int_n);
+            freq_lo2 = _rxlo2->set_frequency(freq_lo1 - freq, ref_freq, target_pfd_freq, is_int_n);
             _rxlo2->set_output_power(max287x_iface::OUTPUT_POWER_2DBM);
         }
         else if ((freq >= 100*fMHz) && (freq < 500*fMHz))
@@ -988,10 +1042,10 @@ private:
             set_cpld_field(RXLB_SEL, 1);
             set_cpld_field(RXHB_SEL, 0);
             // Set LO1 to IF of 2440 (center of filter)
-            freq_lo1 = _rxlo1->set_frequency(2440*fMHz, ref_freq, _rx_target_pfd_freq, is_int_n);
+            freq_lo1 = _rxlo1->set_frequency(2440*fMHz, ref_freq, target_pfd_freq, is_int_n);
             _rxlo1->set_output_power(max287x_iface::OUTPUT_POWER_5DBM);
             // Set LO2 to IF minus desired frequency
-            freq_lo2 = _rxlo2->set_frequency(freq_lo1 - freq, ref_freq, _rx_target_pfd_freq, is_int_n);
+            freq_lo2 = _rxlo2->set_frequency(freq_lo1 - freq, ref_freq, target_pfd_freq, is_int_n);
             _rxlo1->set_output_power(max287x_iface::OUTPUT_POWER_2DBM);
         }
         else if ((freq >= 500*fMHz) && (freq < 800*fMHz))
@@ -1003,7 +1057,7 @@ private:
             set_cpld_field(RXLO1_FSEL1, 1);
             set_cpld_field(RXLB_SEL, 0);
             set_cpld_field(RXHB_SEL, 1);
-            freq_lo1 = _rxlo1->set_frequency(freq, ref_freq, _rx_target_pfd_freq, is_int_n);
+            freq_lo1 = _rxlo1->set_frequency(freq, ref_freq, target_pfd_freq, is_int_n);
             _rxlo1->set_output_power(max287x_iface::OUTPUT_POWER_2DBM);
         }
         else if ((freq >= 800*fMHz) && (freq < 1000*fMHz))
@@ -1015,7 +1069,7 @@ private:
             set_cpld_field(RXLO1_FSEL1, 1);
             set_cpld_field(RXLB_SEL, 0);
             set_cpld_field(RXHB_SEL, 1);
-            freq_lo1 = _rxlo1->set_frequency(freq, ref_freq, _rx_target_pfd_freq, is_int_n);
+            freq_lo1 = _rxlo1->set_frequency(freq, ref_freq, target_pfd_freq, is_int_n);
             _rxlo1->set_output_power(max287x_iface::OUTPUT_POWER_5DBM);
         }
         else if ((freq >= 1000*fMHz) && (freq < 1500*fMHz))
@@ -1027,7 +1081,7 @@ private:
             set_cpld_field(RXLO1_FSEL1, 0);
             set_cpld_field(RXLB_SEL, 0);
             set_cpld_field(RXHB_SEL, 1);
-            freq_lo1 = _rxlo1->set_frequency(freq, ref_freq, _rx_target_pfd_freq, is_int_n);
+            freq_lo1 = _rxlo1->set_frequency(freq, ref_freq, target_pfd_freq, is_int_n);
             _rxlo1->set_output_power(max287x_iface::OUTPUT_POWER_2DBM);
         }
         else if ((freq >= 1500*fMHz) && (freq < 2200*fMHz))
@@ -1039,7 +1093,7 @@ private:
             set_cpld_field(RXLO1_FSEL1, 0);
             set_cpld_field(RXLB_SEL, 0);
             set_cpld_field(RXHB_SEL, 1);
-            freq_lo1 = _rxlo1->set_frequency(freq, ref_freq, _rx_target_pfd_freq, is_int_n);
+            freq_lo1 = _rxlo1->set_frequency(freq, ref_freq, target_pfd_freq, is_int_n);
             _rxlo1->set_output_power(max287x_iface::OUTPUT_POWER_2DBM);
         }
         else if ((freq >= 2200*fMHz) && (freq < 2500*fMHz))
@@ -1051,7 +1105,7 @@ private:
             set_cpld_field(RXLO1_FSEL1, 0);
             set_cpld_field(RXLB_SEL, 0);
             set_cpld_field(RXHB_SEL, 1);
-            freq_lo1 = _rxlo1->set_frequency(freq, ref_freq, _rx_target_pfd_freq, is_int_n);
+            freq_lo1 = _rxlo1->set_frequency(freq, ref_freq, target_pfd_freq, is_int_n);
             _rxlo1->set_output_power(max287x_iface::OUTPUT_POWER_2DBM);
         }
         else if ((freq >= 2500*fMHz) && (freq <= 6000*fMHz))
@@ -1063,7 +1117,7 @@ private:
             set_cpld_field(RXLO1_FSEL1, 0);
             set_cpld_field(RXLB_SEL, 0);
             set_cpld_field(RXHB_SEL, 1);
-            freq_lo1 = _rxlo1->set_frequency(freq, ref_freq, _rx_target_pfd_freq, is_int_n);
+            freq_lo1 = _rxlo1->set_frequency(freq, ref_freq, target_pfd_freq, is_int_n);
             _rxlo1->set_output_power(max287x_iface::OUTPUT_POWER_5DBM);
         }
 
@@ -1098,7 +1152,7 @@ private:
         _rxlo1_freq = freq_lo1;
         _rxlo2_freq = freq_lo2;
 
-        UHD_LOGV(rarely) << boost::format("UBX RX: the actual frequency is %f MHz") % (_rx_freq/1e6) << std::endl;
+        UHD_LOGGER_TRACE("UBX") << boost::format("UBX RX: the actual frequency is %f MHz") % (_rx_freq/1e6) ;
 
         return _rx_freq;
     }
@@ -1125,7 +1179,7 @@ private:
 
             // Placeholders in case some components need to be forced on to
             // reduce settling time.  Note that some FORCEON lines are still gated
-            // by other bits in the CPLD register are are asserted during
+            // by other bits in the CPLD register and are asserted during
             // frequency tuning.
             set_cpld_field(RXAMP_FORCEON, 1);
             set_cpld_field(RXDEMOD_FORCEON, 1);
@@ -1181,7 +1235,29 @@ private:
         // The intent is to add behavior based on whether
         // the board is in TX, RX, or full duplex mode
         // to reduce power consumption and RF noise.
-        _xcvr_mode = mode;
+        boost::to_upper(mode);
+        if (mode == "FDX")
+        {
+            _xcvr_mode = FDX;
+        }
+        else if (mode == "TDD")
+        {
+            _xcvr_mode = TDD;
+            set_cpld_field(TXDRV_FORCEON, 1);
+            write_cpld_reg();
+        }
+        else if (mode == "TX")
+        {
+            _xcvr_mode = TX;
+        }
+        else if (mode == "RX")
+        {
+            _xcvr_mode = RX;
+        }
+        else
+        {
+            throw uhd::value_error("invalid xcvr_mode");
+        }
     }
 
     void set_sync_delay(bool is_tx, int64_t value)
@@ -1221,14 +1297,13 @@ private:
     int _ubx_tx_atten_val;
     int _ubx_rx_atten_val;
     power_mode_t _power_mode;
-    std::string _xcvr_mode;
+    xcvr_mode_t _xcvr_mode;
     size_t _rev;
-    double _prev_tx_freq;
-    double _prev_rx_freq;
     ubx_gpio_reg_t _tx_gpio_reg;
     ubx_gpio_reg_t _rx_gpio_reg;
     int64_t _tx_sync_delay;
     int64_t _rx_sync_delay;
+    bool _high_isolation;
 };
 
 /***********************************************************************
@@ -1248,4 +1323,5 @@ UHD_STATIC_BLOCK(reg_ubx_dboards)
     dboard_manager::register_dboard(UBX_V2_40MHZ_RX_ID,  UBX_V2_40MHZ_TX_ID,  &make_ubx, "UBX-40 v2");
     dboard_manager::register_dboard(UBX_V2_160MHZ_RX_ID, UBX_V2_160MHZ_TX_ID, &make_ubx, "UBX-160 v2");
     dboard_manager::register_dboard(UBX_LP_160MHZ_RX_ID, UBX_LP_160MHZ_TX_ID, &make_ubx, "UBX-160-LP");
+    dboard_manager::register_dboard(UBX_TDD_160MHZ_RX_ID, UBX_TDD_160MHZ_TX_ID, &make_ubx, "UBX-TDD");
 }
